const DEFAULT_BRIDGE_TRANSFER_OPTIONS = {
    release: true,
    transfer: false
};
class BridgeSerializer {
    constructor(serializer) {
        this.serializer = serializer;
    }
    serialize(copy, wrap) {
        return this.serializer(copy, wrap);
    }
}
class PausedBridge {
    call(fn, ...args) {
        throw new Error(PausedBridge.NOT_BOOTSTRAPPED_ERROR(fn));
    }
    callIgnored(fn, ...args) {
        throw new Error(PausedBridge.NOT_BOOTSTRAPPED_ERROR(fn));
    }
    callSync(fn, ...args) {
        throw new Error(PausedBridge.NOT_BOOTSTRAPPED_ERROR(fn));
    }
}
PausedBridge.NOT_BOOTSTRAPPED_ERROR = (fn) => `Cannot call: ${fn} before bootstrapping runtime`;
class ActiveBridge {
    constructor(dispatch, ivm) {
        this.dispatch = dispatch;
        this.ivm = ivm;
    }
    externalCopy(arg) {
        return new this.ivm.ExternalCopy(arg);
    }
    call(fn, ...args) {
        return this.apply(this.dispatch, fn, args);
    }
    callIgnored(fn, ...args) {
        return this.applyIgnored(this.dispatch, fn, args);
    }
    callSync(fn, ...args) {
        return this.applySync(this.dispatch, fn, args);
    }
    apply(receiver, fn, args) {
        return receiver.apply(undefined, [fn, ...args.map(a => this.wrap(a))]);
    }
    applySync(receiver, fn, args) {
        return receiver.applySync(undefined, [fn, ...args.map(a => this.wrap(a))]);
    }
    applyIgnored(receiver, ...args) {
        return receiver.applyIgnored(undefined, args.map(a => this.wrap(a)));
    }
    wrap(arg, opts = {}) {
        if (!arg) {
            return arg;
        }
        if (arg instanceof Error) {
            const transformedErrorObj = {
                name: arg.name || '',
                message: arg.message || '',
                stack: arg.stack || ''
            };
            return this.wrapValue(transformedErrorObj, { release: !!opts.release, transfer: false });
        }
        if (arg instanceof BridgeSerializer) {
            arg = arg.serialize((arg) => this.externalCopy(arg), (fn, opts) => this.wrapFunction(fn, opts));
        }
        let ctor;
        try {
            ctor = arg.constructor;
        }
        catch (e) {
            ctor = undefined;
        }
        switch (ctor) {
            case this.ivm.ExternalCopy:
            case this.ivm.Reference:
            case String:
            case Number:
            case Boolean:
                return arg;
            case ArrayBuffer:
            case Uint8Array:
            case Int8Array:
            case Uint8ClampedArray:
            case Int16Array:
            case Uint16Array:
            case Int32Array:
            case Uint32Array:
            case Float32Array:
            case Float64Array:
                return this.wrapValue(arg, opts);
            case DataView:
            case Map:
            case Set:
            case RegExp:
            case Date:
            case Object:
                return this.wrapValue(arg, { release: !!opts.release, transfer: false });
            case Array:
                return this.wrapArray(arg, { release: !!opts.release, transfer: false });
            case Function:
                return this.wrapFunction(arg);
            default:
                throw new Error(`Can't wrap a non-transferable value (constructor: '${(ctor && ctor.name) || "unknown"}')`);
        }
    }
    wrapFunction(fn, options = { release: true }) {
        const opts = Object.assign({}, DEFAULT_BRIDGE_TRANSFER_OPTIONS, options || {});
        if (!opts.release) {
            return new this.ivm.Reference(fn);
        }
        const cb = new this.ivm.Reference(function bridgeAutoReleaseFn(...args) {
            try {
                cb.release();
            }
            catch (err) {
            }
            finally {
                fn(...args);
            }
        });
        return cb;
    }
    wrapValue(value, options = { release: true }) {
        const opts = Object.assign({}, DEFAULT_BRIDGE_TRANSFER_OPTIONS, options || {});
        if (!!opts.transfer) {
            return new this.ivm.ExternalCopy(value, { transferOut: true }).copyInto({
                release: !!opts.release,
                transferIn: true
            });
        }
        return new this.ivm.ExternalCopy(value).copyInto({ release: !!opts.release });
    }
    wrapArray(value, options = { release: true }) {
        return this.wrapValue(value.map(v => this.wrap(v)), options);
    }
}
let bridge = new PausedBridge();
(() => {
    const isTransferableThroughBridge = (obj) => {
        switch (obj.constructor) {
            case String:
            case Number:
            case Boolean:
            case ArrayBuffer:
            case Uint8Array:
            case Int8Array:
            case Uint8ClampedArray:
            case Int16Array:
            case Uint16Array:
            case Int32Array:
            case Uint32Array:
            case Float32Array:
            case Float64Array:
            case DataView:
            case Map:
            case Set:
            case RegExp:
            case Date:
            case Object:
            case Array:
                return true;
            default:
                return false;
        }
        ;
    };
    const transformNonTransferablesForLogging = (externalCopy, object) => {
        var visitedObjects = [];
        const isObject = (obj) => obj && typeof obj === 'object';
        const isPromise = (obj) => obj.constructor === Promise;
        const isObjectLiteral = (obj) => obj.constructor === Object;
        const isNullPrototypeObject = (obj) => Object.getPrototypeOf(obj) === null;
        const getTag = (obj) => {
            const desc = Object.getOwnPropertyDescriptor(obj, Symbol.toStringTag);
            if (desc && desc.enumerable === false) {
                return `[${desc.value}]`;
            }
            return '';
        };
        const serializeToExternalCopy = (value, prefix) => {
            return externalCopy({
                value,
                prefix
            });
        };
        function _transform(value) {
            const formattedObject = (constructor, tag, obj) => {
                visitedObjects.push(obj);
                const enumerableProps = _transform(Object.assign({}, obj));
                visitedObjects.pop();
                const val = [constructor, tag].filter(x => x !== '').join(' ');
                return serializeToExternalCopy(enumerableProps, val);
            };
            if (typeof value === 'function') {
                return serializeToExternalCopy(`[Function${value.name ? ': ' + value.name : ''}]`, undefined);
            }
            if (typeof value === 'symbol') {
                return serializeToExternalCopy(value.toString(), undefined);
            }
            if (isObject(value)) {
                const tag = getTag(value);
                if (visitedObjects.indexOf(value) > -1) {
                    return serializeToExternalCopy('[Circular]', undefined);
                }
                else if (Array.isArray(value)) {
                    value = value.map(_transform);
                }
                else if (isPromise(value)) {
                    return serializeToExternalCopy('[Promise]', undefined);
                }
                else if (isNullPrototypeObject(value)) {
                    return formattedObject('[Object: null prototype]', tag, value);
                }
                else if (isObjectLiteral(value)) {
                    if (tag) {
                        return formattedObject('Object', tag, value);
                    }
                    visitedObjects.push(value);
                    const newValue = {};
                    Object.entries(value).forEach(([key, val]) => newValue[key] = _transform(val));
                    Object.getOwnPropertySymbols(value).forEach((symbolKey) => newValue[`[${symbolKey.toString()}]`] = _transform(value[symbolKey]));
                    value = newValue;
                    visitedObjects.pop();
                }
                else if (!isTransferableThroughBridge(value)) {
                    const name = value.constructor ? value.constructor.name : '';
                    return formattedObject(name, tag, value);
                }
            }
            return value;
        }
        return _transform(object);
    };
    const logConsole = (method) => (...args) => {
        bridge.call('logConsole', method, new BridgeSerializer((externalCopy, _) => (args.map(arg => transformNonTransferablesForLogging(externalCopy, arg)))));
    };
    global.console = {
        debug: logConsole('debug'),
        error: logConsole('error'),
        warn: logConsole('warn'),
        info: logConsole('info'),
        log: logConsole('log'),
    };
    global.setTimeout = (callback, ms, ...args) => {
        return bridge.callSync('setTimeout', new BridgeSerializer((_, wrapFunction) => ({
            callback: wrapFunction(() => callback(...args)),
            ms
        })));
    };
    global.clearTimeout = (id) => {
        return bridge.callSync('clearTimeout', () => ({ id }));
    };
    global.setInterval = (callback, ms, ...args) => {
        return bridge.callSync('setInterval', new BridgeSerializer((_, wrapFunction) => ({
            callback: wrapFunction(() => callback(...args)),
            ms
        })));
    };
    global.clearInterval = (id) => {
        return bridge.callSync('clearInterval', () => ({ id }));
    };
    function apiFetch(bridgeMethodToCall, url, init) {
        return new Promise((resolve, reject) => {
            const resolveWrapper = (response) => {
                const fetchResponse = {
                    text: async () => response.data,
                    json: async () => JSON.parse(response.data),
                    arrayBuffer: async () => response.arrayBufferData,
                    ok: response.ok,
                    status: response.status,
                    statusText: response.statusText
                };
                resolve(fetchResponse);
            };
            const rejectWrapper = (response) => {
                if (response.error)
                    reject(response.error);
                reject(response);
            };
            bridge.callIgnored(bridgeMethodToCall, url, init, resolveWrapper, rejectWrapper);
        });
    }
    function requestJira(url, init) {
        return apiFetch('api.requestJira', url, init);
    }
    function requestConfluence(url, init) {
        return apiFetch('api.requestConfluence', url, init);
    }
    function fetch(url, init) {
        return apiFetch('api.fetch', url, init);
    }
    function asUser() {
        return {
            requestJira: (url, init) => apiFetch('api.asUser.requestJira', url, init),
            requestConfluence: (url, init) => apiFetch('api.asUser.requestConfluence', url, init),
        };
    }
    function asApp() {
        return {
            requestJira: (url, init) => apiFetch('api.asApp.requestJira', url, init),
            requestConfluence: (url, init) => apiFetch('api.asApp.requestConfluence', url, init),
        };
    }
    global.api = {
        asUser,
        asApp,
        fetch,
        requestJira,
        requestConfluence,
    };
})();
global.bootstrap = (ivm, dispatch, FFisMultiAuthEnabled) => {
    delete global.bootstrap;
    const activeBridge = new ActiveBridge(dispatch, ivm);
    bridge = activeBridge;
    const createIOStream = (fd) => ({ fd });
    function Process() { }
    Object.defineProperty(Process.prototype, Symbol.toStringTag, {
        get: function () {
            return "process";
        }
    });
    global.process = {
        env: JSON.parse(bridge.callSync('getUserVars')),
        platform: bridge.callSync('getUserPlatform'),
        version: bridge.callSync('getUserVersion'),
        versions: bridge.callSync('getUserVersions'),
        nextTick: (callback, ...args) => {
            setTimeout(callback, 0, ...args);
        },
        stdin: createIOStream(0),
        stdout: createIOStream(1),
        stderr: createIOStream(2),
    };
    Object.setPrototypeOf(global.process, Process.prototype);
    function getAuthTokenReferences(serviceKey) {
        return bridge.callSync('api.getAuthTokenReferences', serviceKey);
    }
    if (FFisMultiAuthEnabled)
        global.api.getAuthTokenReferences = getAuthTokenReferences;
    Object.defineProperty(global, '__invoke', {
        configurable: false,
        enumerable: false,
        writable: false,
        value: (fn, args, context, resolve, reject) => {
            const scopedFn = global[fn] || exports[fn];
            if (!scopedFn) {
                throw new Error(`Can't invoke a function "${fn}". Maybe it's not exported?`);
            }
            const result = scopedFn.apply(undefined, [args, context]);
            Promise.resolve(result)
                .then((result) => {
                activeBridge.applyIgnored(resolve, result);
            }).catch(err => {
                activeBridge.applyIgnored(reject, err);
            });
        }
    });
    Object.defineProperty(global, '__bridge', {
        configurable: false,
        enumerable: false,
        writable: false,
        value: bridge
    });
};
